@@ -9,6 +9,7 @@ Multi-line commands
 Case-insensitive commands
 Special-character shortcut commands (beyond cmd's "@" and "!")
 Settable environment parameters
+Optional _onchange_{paramname} called when environment parameter changes
 Parsing commands with `optparse` options (flags)
 Redirection to file with >, >>; input from file with <
 Easy transcript-based testing of applications (see example/example.py)
@@ -25,15 +26,23 @@ As of 0.3.0, options should be specified as `optparse` options.  See README.txt.
 flagReader.py options are still supported for backward compatibility
 """
 import cmd, re, os, sys, optparse, subprocess, tempfile, pyparsing, doctest
-import unittest, string, datetime
+import unittest, string, datetime, urllib
 from optparse import make_option
-__version__ = '0.4.7'
+__version__ = '0.4.8'
 
 class OptionParser(optparse.OptionParser):
     def exit(self, status=0, msg=None):
         self.values._exit = True
         if msg:
             print msg
+            
+    def print_help(self, *args, **kwargs):
+        # now, I need to call help of the calling function.  Hmm.
+        try:
+            print self._func.__doc__
+        except AttributeError:
+            pass
+        optparse.OptionParser.print_help(self, *args, **kwargs)
 
     def error(self, msg):
         """error(msg : string)
@@ -59,6 +68,7 @@ def options(option_list):
         for opt in option_list:
             optionParser.add_option(opt)
         optionParser.set_usage("%s [options] arg" % func.__name__.strip('do_'))
+        optionParser._func = func
         def newFunc(instance, arg):
             try:
                 opts, newArgList = optionParser.parse_args(arg.split()) # doesn't understand quoted strings shouldn't be dissected!
@@ -71,13 +81,17 @@ def options(option_list):
                 return
             if hasattr(opts, '_exit'):
                 return None
-            terminator = arg.parsed.terminator
-            try:
-                if arg.parsed.terminator[0] == '\n':
-                    terminator = arg.parsed.terminator[0]
-            except IndexError:
-                pass
-            arg = arg.parser('%s %s%s%s' % (arg.parsed.command, newArgs, terminator, arg.parsed.suffix))
+            if hasattr(arg, 'parser'):
+                terminator = arg.parsed.terminator
+                try:
+                    if arg.parsed.terminator[0] == '\n':
+                        terminator = arg.parsed.terminator[0]
+                except IndexError:
+                    pass
+                arg = arg.parser('%s %s%s%s' % (arg.parsed.command, newArgs, 
+                                                terminator, arg.parsed.suffix))
+            else:
+                arg = newArgs
             result = func(instance, arg, opts)                            
             return result        
         newFunc.__doc__ = '%s\n%s' % (func.__doc__, optionParser.format_help())
@@ -199,7 +213,6 @@ class SkipToLast(pyparsing.SkipTo):
             raise exc    
 
 def replace_with_file_contents(fname):
-    import pdb; pdb.set_trace()
     if fname:
         try:
             result = open(os.path.expanduser(fname[0])).read()
@@ -211,16 +224,16 @@ def replace_with_file_contents(fname):
         
 class Cmd(cmd.Cmd):
     echo = False
-    caseInsensitive = True
-    continuationPrompt = '> '  
+    case_insensitive = True
+    continuation_prompt = '> '  
     timing = False
     legalChars = '!#$%.:?@_' + pyparsing.alphanums + pyparsing.alphas8bit  # make sure your terminators are not in here!
     shortcuts = {'?': 'help', '!': 'shell', '@': 'load' }
     excludeFromHistory = '''run r list l history hi ed edit li eof'''.split()
     noSpecialParse = 'set ed edit exit'.split()
     defaultExtension = 'txt'
-    defaultFileName = 'command.txt'
-    settable = ['prompt', 'continuationPrompt', 'defaultFileName', 'editor', 'caseInsensitive', 
+    default_file_name = 'command.txt'
+    settable = ['prompt', 'continuation_prompt', 'default_file_name', 'editor', 'case_insensitive', 
                 'echo', 'timing']
     settable.sort()
     
@@ -241,30 +254,33 @@ class Cmd(cmd.Cmd):
         Commands may be terminated with: %(terminators)s
         Settable parameters: %(settable)s
         """ % 
-        { 'casesensitive': (self.caseInsensitive and 'not ') or '',
+        { 'casesensitive': (self.case_insensitive and 'not ') or '',
           'terminators': str(self.terminators),
           'settable': ' '.join(self.settable)
         })
         
     def do_help(self, arg):
-        cmd.Cmd.do_help(self, arg)
         try:
             fn = getattr(self, 'do_' + arg)
             if fn and fn.optionParser:
                 fn.optionParser.print_help(file=self.stdout)
+                return
         except AttributeError:
             pass
+        cmd.Cmd.do_help(self, arg)
         
     def __init__(self, *args, **kwargs):        
         cmd.Cmd.__init__(self, *args, **kwargs)
         self.history = History()
         self._init_parser()
+        self.pystate = {}
         
     def do_shortcuts(self, args):
         """Lists single-key shortcuts available."""
         result = "\n".join('%s: %s' % (sc[0], sc[1]) for sc in self.shortcuts.items())
         self.stdout.write("Single-key shortcuts for other commands:\n%s\n" % (result))
 
+    prefixParser = pyparsing.Empty()
     commentGrammars = pyparsing.Or([pyparsing.pythonStyleComment, pyparsing.cStyleComment])
     commentGrammars.addParseAction(lambda x: '')
     commentInProgress  = pyparsing.Literal('/*') + pyparsing.SkipTo(pyparsing.stringEnd)
@@ -276,7 +292,7 @@ class Cmd(cmd.Cmd):
         r'''
         >>> c = Cmd()
         >>> c.multilineCommands = ['multiline']
-        >>> c.caseInsensitive = True
+        >>> c.case_insensitive = True
         >>> c._init_parser()
         >>> print c.parser.parseString('').dump()        
         []        
@@ -434,7 +450,7 @@ class Cmd(cmd.Cmd):
         
         terminatorParser = pyparsing.Or([(hasattr(t, 'parseString') and t) or pyparsing.Literal(t) for t in self.terminators])('terminator')
         stringEnd = pyparsing.stringEnd ^ '\nEOF'
-        self.multilineCommand = pyparsing.Or([pyparsing.Keyword(c, caseless=self.caseInsensitive) for c in self.multilineCommands])('multilineCommand')
+        self.multilineCommand = pyparsing.Or([pyparsing.Keyword(c, caseless=self.case_insensitive) for c in self.multilineCommands])('multilineCommand')
         oneLineCommand = (~self.multilineCommand + pyparsing.Word(self.legalChars))('command')
         pipe = pyparsing.Keyword('|', identChars='|')
         self.commentGrammars.ignore(pyparsing.quotedString).setParseAction(lambda x: '')
@@ -442,7 +458,7 @@ class Cmd(cmd.Cmd):
         afterElements = \
             pyparsing.Optional(pipe + pyparsing.SkipTo(outputParser ^ stringEnd)('pipeTo')) + \
             pyparsing.Optional(outputParser + pyparsing.SkipTo(stringEnd).setParseAction(lambda x: x[0].strip())('outputTo'))
-        if self.caseInsensitive:
+        if self.case_insensitive:
             self.multilineCommand.setParseAction(lambda x: x[0].lower())
             oneLineCommand.setParseAction(lambda x: x[0].lower())
         if self.blankLinesAllowed:
@@ -460,10 +476,10 @@ class Cmd(cmd.Cmd):
         #self.blankLineTerminationParser = self.blankLineTerminationParser.setResultsName('blankLineTerminatorParser')
         self.parser = (
             stringEnd |
-            self.multilineParser |
-            self.singleLineParser |
-            self.blankLineTerminationParser | 
-            self.multilineCommand + pyparsing.SkipTo(stringEnd)
+            self.prefixParser + self.multilineParser |
+            self.prefixParser + self.singleLineParser |
+            self.prefixParser + self.blankLineTerminationParser | 
+            self.prefixParser + self.multilineCommand + pyparsing.SkipTo(stringEnd)
             )
         self.parser.ignore(pyparsing.quotedString).ignore(self.commentGrammars).ignore(self.commentInProgress)
         
@@ -478,10 +494,14 @@ class Cmd(cmd.Cmd):
                            pyparsing.Optional(fileName) + (pyparsing.stringEnd | '|')
         self.inputParser.ignore(pyparsing.quotedString).ignore(self.commentGrammars).ignore(self.commentInProgress)               
     
+    def preparse(self, raw, **kwargs):
+        return raw
+    
     def parsed(self, raw, **kwargs):
         if isinstance(raw, ParsedString):
             p = raw
         else:
+            raw = self.preparse(raw, **kwargs)
             s = self.inputParser.transformString(raw.lstrip())
             for (shortcut, expansion) in self.shortcuts.items():
                 if s.lower().startswith(shortcut):
@@ -499,6 +519,11 @@ class Cmd(cmd.Cmd):
             p.parsed[key] = val
         return p
               
+    def postparsing_precmd(self, statement):
+        stop = 0
+        return stop, statement
+    def postparsing_postcmd(self, stop):
+        return stop
     def onecmd(self, line):
         """Interpret the argument as though it had been typed in response
         to the prompt.
@@ -519,17 +544,24 @@ class Cmd(cmd.Cmd):
             statement = self.parsed(line)
             while statement.parsed.multilineCommand and (statement.parsed.terminator == ''):
                 statement = '%s\n%s' % (statement.parsed.raw, 
-                                        self.pseudo_raw_input(self.continuationPrompt))                
+                                        self.pseudo_raw_input(self.continuation_prompt))                
                 statement = self.parsed(statement)
         except Exception, e:
             print e
             return 0
 
-        if not statement.parsed.command:
+        try:
+            (stop, statement) = self.postparsing_precmd(statement)
+        except Exception, e:
+            print str(e)
             return 0
+        if stop:
+            return self.postparsing_postcmd(stop)
         
+        if not statement.parsed.command:
+            return self.postparsing_postcmd(stop=0)
+                
         statekeeper = None
-        stop = 0
 
         if statement.parsed.pipeTo:
             redirect = subprocess.Popen(statement.parsed.pipeTo, shell=True, stdout=subprocess.PIPE, stdin=subprocess.PIPE)
@@ -545,7 +577,7 @@ class Cmd(cmd.Cmd):
                     self.stdout = open(os.path.expanduser(statement.parsed.outputTo), mode)                            
                 except OSError, e:
                     print e
-                    return 0                    
+                    return self.postparsing_postcmd(stop=0) 
             else:
                 statekeeper = Statekeeper(self, ('stdout',))
                 self.stdout = tempfile.TemporaryFile()
@@ -557,7 +589,7 @@ class Cmd(cmd.Cmd):
             try:
                 func = getattr(self, 'do_' + statement.parsed.command)
             except AttributeError:
-                return self.default(statement)
+                return self.postparsing_postcmd(self.default(statement))
             timestart = datetime.datetime.now()
             stop = func(statement) 
             if self.timing:
@@ -581,7 +613,7 @@ class Cmd(cmd.Cmd):
                 self.stdout.close()
                 statekeeper.restore()
                                  
-            return stop        
+            return self.postparsing_postcmd(stop)        
         
     def pseudo_raw_input(self, prompt):
         """copied from cmd's cmdloop; like raw_input, but accounts for changed stdin, stdout"""
@@ -651,11 +683,15 @@ class Cmd(cmd.Cmd):
     do_eof = do_EOF
                
     def showParam(self, param):
+        any_shown = False
         param = param.strip().lower()
         for p in self.settable:
             if p.startswith(param):
                 val = getattr(self, p)
                 self.stdout.write('%s: %s\n' % (p, str(getattr(self, p))))
+                any_shown = True
+        if not any_shown:
+            print "Parameter '%s' not supported (type 'show' for list of parameters)." % param
                 
     def do_quit(self, arg):
         return self._STOP_AND_EXIT
@@ -663,7 +699,7 @@ class Cmd(cmd.Cmd):
     do_q = do_quit
     
     def do_show(self, arg):
-        'Shows value of a parameter'
+        '''Shows value of a parameter.'''
         if arg.strip():
             self.showParam(arg)
         else:
@@ -686,6 +722,12 @@ class Cmd(cmd.Cmd):
                     val = cast(currentVal, val)
                 setattr(self, paramName, val)
                 self.stdout.write('%s - was: %s\nnow: %s\n' % (paramName, currentVal, val))
+                if currentVal != val:
+                    try:
+                        onchange_hook = getattr(self, '_onchange_%s' % paramName)
+                        onchange_hook(old=currentVal, new=val)
+                    except AttributeError:
+                        pass
             else:
                 self.do_show(paramName)
         except (ValueError, AttributeError, NotSettableError), e:
@@ -699,6 +741,40 @@ class Cmd(cmd.Cmd):
         'execute a command as if at the OS prompt.'
         os.system(arg)
         
+    def do_py(self, arg):  
+        '''
+        py <command>: Executes a Python command.
+        py: Enters interactive Python mode (end with `end py`).
+        '''
+        if arg.strip():
+            try:
+                result = eval(arg, self.pystate)
+                print repr(result)
+            except SyntaxError:
+                try:
+                    exec(arg, self.pystate)
+                except Exception:
+                    raise
+            except Exception, e:
+                print e        
+        else:
+            print 'Now accepting python commands; end with `end py`'
+            buffer = [self.pseudo_raw_input('>>> ')]
+            while buffer[-1].lower().split()[:2] != ['end','py']:
+                try:
+                    buf = '\n'.join(buffer)
+                    try:
+                        result = eval(buf, self.pystate)
+                        print repr(result)
+                    except SyntaxError:
+                        exec(buf, self.pystate)
+                    buffer = [self.pseudo_raw_input('>>> ')]
+                except SyntaxError:
+                    buffer.append(self.pseudo_raw_input('... '))
+                except Exception, e:
+                    print e
+                    buffer = [self.pseudo_raw_input('>>> ')]
+
     def do_history(self, arg):
         """history [arg]: lists past commands issued
         
@@ -749,7 +825,7 @@ class Cmd(cmd.Cmd):
         if not self.editor:
             print "please use 'set editor' to specify your text editing program of choice."
             return
-        filename = self.defaultFileName
+        filename = self.default_file_name
         if arg:
             try:
                 buffer = self.last_matching(int(arg))
@@ -782,7 +858,7 @@ class Cmd(cmd.Cmd):
         except pyparsing.ParseException:
             print self.do_save.__doc__
             return
-        fname = args.fname or self.defaultFileName
+        fname = args.fname or self.default_file_name
         if args.idx == '*':
             saveme = '\n\n'.join(self.history[:])
         elif args.idx:
@@ -797,26 +873,34 @@ class Cmd(cmd.Cmd):
         except Exception, e:
             print 'Error saving %s: %s' % (fname, str(e))
             
-    def do_load(self, fname=None):
-        """Runs command(s) from a file."""
+    urlre = re.compile('(https?://[-\\w\\./]+)')
+    def do_load(self, fname=None):           
+        """Runs script of command(s) from a file or URL."""
         if fname is None:
-            fname = self.defaultFileName
-        fname = os.path.expanduser(fname)
-        keepstate = Statekeeper(self, ('stdin','use_rawinput','prompt','continuationPrompt'))
-        if isinstance(fname, file):
-            self.stdin = fname
-        else:           
-            try:
-                self.stdin = open(os.path.expanduser(fname), 'r')
-            except IOError, e:
-                try:
-                    self.stdin = open('%s.%s' % (os.path.expanduser(fname), self.defaultExtension), 'r')
-                except IOError:
-                    print 'Problem opening file %s: \n%s' % (fname, e)
-                    keepstate.restore()
-                    return
+            fname = self.default_file_name
+        #keepstate = Statekeeper(self, ('stdin','use_rawinput','prompt','continuation_prompt'))
+        keepstate = Statekeeper(self, ('stdin','use_rawinput','continuation_prompt'))
+        try:
+            if isinstance(fname, file):
+                target = open(fname, 'r')
+            else:
+                match = self.urlre.match(fname)
+                if match:
+                    target = urllib.urlopen(match.group(1))
+                else:
+                    fname = os.path.expanduser(fname)
+                    try:
+                        target = open(os.path.expanduser(fname), 'r')
+                    except IOError, e:                    
+                        target = open('%s.%s' % (os.path.expanduser(fname), 
+                                                 self.defaultExtension), 'r')
+        except IOError, e:
+            print 'Problem accessing script from %s: \n%s' % (fname, e)
+            keepstate.restore()
+            return
+        self.stdin = target    
         self.use_rawinput = False
-        self.prompt = self.continuationPrompt = ''
+        self.prompt = self.continuation_prompt = ''
         stop = self.cmdloop()
         self.stdin.close()
         keepstate.restore()
@@ -988,8 +1072,8 @@ class Cmd2TestCase(unittest.TestCase):
                         line = self.transcript.next()
                     command = [line[len(self.cmdapp.prompt):]]
                     line = self.transcript.next()
-                    while line.startswith(self.cmdapp.continuationPrompt):
-                        command.append(line[len(self.cmdapp.continuationPrompt):])
+                    while line.startswith(self.cmdapp.continuation_prompt):
+                        command.append(line[len(self.cmdapp.continuation_prompt):])
                         line = self.transcript.next()
                     command = ''.join(command)
                     self.cmdapp.onecmd(command)
